﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Exam2
{
    class Program
    {
        static void Main(string[] args)
        {
            GenerateQ1LatinSquareTestData();
        }

        private static void GenerateQ1LatinSquareTestData()
        {
            int testCaseNo = 1;
            for (int dim=3; dim<=20; dim++)
            {
                var table = GenCorrectTable(dim: dim);
                ConsistentShuffleRows(table);
                ConsistentShuffleColumns(table);
                //table.Print();
                //Console.WriteLine();

                int?[,] partialTable = ErasePercent(table, Rnd.Next(20, 50));
                //partialTable.Print();
                WriteTestCase(partialTable, testCaseNo++, sat: true);

                int?[,] rndShuffle1 = RandomShuffle(partialTable, percent: 1);
                //rndShuffle1.Print();
                WriteTestCase(rndShuffle1, testCaseNo++, sat: false);

                int?[,] rndShuffle2 = RandomShuffle(partialTable, percent: 1);
                //rndShuffle1.Print();
                WriteTestCase(rndShuffle2, testCaseNo++, sat: false);

            }
        }

        private static void WriteTestCase(int?[,] partialTable, int testCaseNo, bool sat)
        {
            string inFile = Path.Combine(@"C:\git\sauleh\AD.Graph\Exam2\Exam2Tests\TestData\TD1", $"In_{testCaseNo}.txt");
            string outFile = inFile.Replace("In_", "Out_");

            Console.WriteLine($"Writing {inFile}");

            File.WriteAllText(outFile, sat ? "SATISFIABLE" : "UNSATISFIABLE");
            File.WriteAllLines(inFile,
                new string[] { partialTable.GetLength(0).ToString() }
                .Concat(partialTable.Rows().Select(idx => string.Join(" ", partialTable.GetRow(idx).Select(e => e == null ? "." : e.ToString())))));

        }

        private static int?[,] RandomShuffle(int?[,] partialTable, double percent)
        {
            int dim = partialTable.GetLength(0);
            int?[,] partialTableCopy = partialTable.Clone() as int?[,];
            for (int i=0; i<(partialTable.Length*percent*2); i++)
            {
                (int row, int col) p1 = (Rnd.Next(0, dim - 1), Rnd.Next(0, dim - 1));
                (int row, int col) p2 = (Rnd.Next(0, dim - 1), Rnd.Next(0, dim - 1));

                (partialTableCopy[p1.row, p1.col], partialTableCopy[p2.row, p2.col]) =
                    (partialTableCopy[p2.row, p2.col], partialTableCopy[p1.row, p1.col]);
            }

            // Remove duplicates in rows and columns
            for (int i = 0; i < dim; i++)
                for (int j = 0; j < dim; j++)
                {
                    for (int k = j + 1; k < dim; k++)
                        if (partialTableCopy[i, j].HasValue &&
                            partialTableCopy[i, j] == partialTableCopy[i, k])
                            partialTableCopy[i, k] = null;

                    for (int k= i+1; k<dim; k++)
                        if (partialTableCopy[i, j].HasValue &&
                            partialTableCopy[i, j] == partialTableCopy[k, j])
                            partialTableCopy[k, j] = null;
                }


            return partialTableCopy;
        }

        private static int?[,] ErasePercent(int[,] table, int p)
        {
            int?[,] result = new int?[table.GetLength(0), table.GetLength(1)];
            for (int i = 0; i < table.GetLength(0); i++)
                for (int j = 0; j < table.GetLength(0); j++)
                {
                    if (Rnd.NextDouble() * 100 > p)
                        result[i, j] = table[i, j];
                }
            return result;
        }

        private static void ConsistentShuffleColumns(int[,] table)
        {
            int dim = table.GetLength(0);
            int[] tmp1 = new int[dim];
            int[] tmp2 = new int[dim];
            for (int c = 0; c < ShuffleCount; c++)
            {
                int col1 = Rnd.Next(0, dim - 1);
                int col2 = Rnd.Next(0, dim - 1);
                CopyCol(table, col: col1, dst: tmp1);
                CopyCol(table, col: col2, dst: tmp2);

                CopyIntoCol(tmp1, col2, table);
                CopyIntoCol(tmp2, col1, table);
            }
        }

        static Random Rnd = new Random(0);
        const int ShuffleCount = 20;
        private static void ConsistentShuffleRows(int[,] table)
        {
            int dim = table.GetLength(0);
            int[] tmp1 = new int[dim];
            int[] tmp2 = new int[dim];
            for (int c=0; c<ShuffleCount; c++)
            {
                int row1 = Rnd.Next(0, dim - 1);
                int row2 = Rnd.Next(0, dim - 1);
                CopyRow(table, row: row1, dst: tmp1);
                CopyRow(table, row: row2, dst: tmp2);

                CopyIntoRow(tmp1, row2, table);
                CopyIntoRow(tmp2, row1, table);
            }
        }

        private static void CopyRow(int[,] table, int row, int[] dst)
        {
            for (int i = 0; i < dst.Length; i++)
                dst[i] = table[row, i];
        }

        private static void CopyIntoRow(int[] from, int row, int[,] dstTable)
        {
            for (int i = 0; i < from.Length; i++)
                dstTable[row, i] = from[i];
        }

        private static void CopyIntoCol(int[] from, int col, int[,] dstTable)
        {
            for (int i = 0; i < from.Length; i++)
                dstTable[i, col] = from[i];
        }

        private static void CopyCol(int[,] table, int col, int[] dst)
        {
            for (int i = 0; i < dst.Length; i++)
                dst[i] = table[i, col];
        }

        private static int[,] GenCorrectTable(int dim)
        {
            int[,] table = new int[dim, dim];
            LinkedList<int> l = new LinkedList<int>(Enumerable.Range(0, dim));
            for (int i = 0; i < dim; i++)
            {
                var row = l.Rotate().ToArray();
                for (int j = 0; j < row.Length; j++)
                    table[i, j] = row[j];
            }
            return table;
        }
    }

    public static partial class Ext
    {
        public static LinkedList<T> Rotate<T>(this LinkedList<T> list)
        {
            var last = list.Last;
            list.RemoveLast();
            list.AddFirst(last);
            return list;
        }

        public static void Print<T>(this LinkedList<T> list)
        {
            Console.WriteLine(string.Join(" ", list));
        }

        public static void Print<T>(this T[,] table)
        {
            for (int i = 0; i < table.GetLength(0); i++)
            {
                for (int j = 0; j < table.GetLength(1); j++)
                {
                    Console.Write(table[i, j] + " ");
                }
                Console.WriteLine();
            }
        }

        public static IEnumerable<int> Rows<T>(this T?[,] table) where T : struct =>
            Enumerable.Range(0, table.GetLength(0));

        public static IEnumerable<T?> GetRow<T>(this T?[,] table, int row) where T:struct
        {
            for (int i = 0; i < table.GetLength(1); i++)
                yield return table[row, i];

        }

        public static IEnumerable<T?> GetCol<T>(this T?[,] table, int col) where T : struct
        {
            for (int i = 0; i < table.GetLength(0); i++)
                yield return table[i, col];

        }

        public static void Print<T>(this T?[,] table) where T: struct
        {
            for (int i = 0; i < table.GetLength(0); i++)
            {
                for (int j = 0; j < table.GetLength(1); j++)
                {
                    Console.Write((table[i, j] != null ? table[i, j].Value.ToString() : ".") + " ");
                }
                Console.WriteLine();
            }
        }
    }

}
